"""Command-line interface."""
import click

import sparkmon


@click.command()
@click.version_option()
@click.option(
    "-i",
    "--index",
    default=0,
    show_default=True,
    help="Application index or row number.",
)
@click.option("-u", "--url", default=sparkmon.WEB_URL, show_default=True, help="Spark UI web URL.")
@click.option("-p", "--period", default=5, show_default=True, help="Update period in seconds.")
def main(index: int, url: str, period: int) -> None:
    """Command line interface to launch sparkmon, it will monitor your Spark Application.

    The default callback is saving the memoring graph in ./sparkmon.png
    """
    application = sparkmon.create_application_from_link(index, url)

    mon = sparkmon.SparkMon(application, period=period, callbacks=[sparkmon.callbacks.plot_to_image])
    mon.start()
    click.confirm("Press Enter to stop...", abort=False, default=True, show_default=False)
    mon.stop()


if __name__ == "__main__":
    main(prog_name="sparkmon")  # pragma: no cover



# noreorder # noqa
from .application import *  # noqa
from .monitor import *  # noqa
from .plotting import *  # noqa

# Needs to be the end:
from . import callbacks  # noqa
from . import mlflow_utils  # noqa



# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Plotting utilities."""
from typing import Any
from typing import Callable

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from IPython import display

import sparkmon
from sparkmon.utils import convert_size


def plot_max_value(ax: matplotlib.axes.Axes, s: Any, string_rep: Callable = lambda x: f"{x:0.2f}") -> None:
    """Helping function to add the maximum value on the right side of a plot."""
    ax.annotate(
        string_rep(s.max()),
        xy=(1, s.max()),
        xytext=(8, 0),
        xycoords=("axes fraction", "data"),
        textcoords="offset points",
    )


def ax_convert_size(ax: matplotlib.axes.Axes):
    """Convert byte size for displaying tick values."""
    ticks = [convert_size(x) for x in ax.get_yticks()]
    ax.set_yticks(
        ax.get_yticks()
    )  # Necessary to remove UserWarning: FixedFormatter should only be used together with FixedLocator
    ax.set_yticklabels(ticks)


def mmm_plot(
    ax: matplotlib.axes.Axes,
    value: str,
    timeseries_db_df: pd.DataFrame,
    title_prefix: str,
    pct: bool = False,
    ylim: Any = (0, None),
) -> None:
    """Helping function to plot the aggregations of a metric."""
    ax.set_title(f"{title_prefix}{value}:")
    if f"{value}_max" not in timeseries_db_df.columns:
        return

    timeseries_db_df[f"{value}_max"].plot(label="max", ax=ax, ylim=ylim)
    timeseries_db_df[f"{value}_median"].plot(label="median", ax=ax, ylim=ylim)
    timeseries_db_df[f"{value}_mean"].plot(label="mean", ax=ax, ylim=ylim)
    timeseries_db_df[f"{value}_min"].plot(label="min", ax=ax, ylim=ylim)
    if not pct:
        ax_convert_size(ax)
    ax.legend()
    plot_max_value(ax, timeseries_db_df[f"{value}_max"], convert_size)


def prepare_axis(ax: matplotlib.axes.Axes) -> None:
    """Prepare subplot ax to have no whitespace before and after the plotlines.

    To call before any plotting.
    """
    ax.margins(x=0)


def plot_timeseries(timeseries_db_df: pd.DataFrame, title: str = None) -> matplotlib.figure.Figure:
    """Plot timeseries DB."""
    if len(timeseries_db_df) == 0:
        return

    fig = plt.figure(constrained_layout=False, figsize=(28, 16), tight_layout={"pad": 0})
    gs = fig.add_gridspec(9, 2)
    if title is not None:
        fig.suptitle(title, y=1)

    ax = fig.add_subplot(gs[0, 0])
    prepare_axis(ax)
    ax.set_title("1. Num active executors:")
    timeseries_db_df["numActive"].plot(ax=ax)
    ax.set_xticklabels([])
    ax.set_ylim(bottom=0)

    ax = fig.add_subplot(gs[0, 1])
    prepare_axis(ax)
    ax.set_title("2. Total memory used percentage:")
    timeseries_db_df["memoryUsed_sum_pct"].plot(ax=ax, ylim=(0, 100))
    plot_max_value(ax, timeseries_db_df["memoryUsed_sum_pct"])
    timeseries_db_df["local_memory_pct"].plot(ax=ax, ylim=(0, 100))
    plot_max_value(ax, timeseries_db_df["local_memory_pct"])
    ax.legend()
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[1:3, 0])
    prepare_axis(ax)
    # mmm_plot(ax, 'memoryUsedPct', timeseries_db_df, '3. ', pct=True, ylim=(0,100))
    mmm_plot(ax, "OnHeapExecutionMemory", timeseries_db_df, "3. Peak ")
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[1:3, 1])
    prepare_axis(ax)
    # mmm_plot(ax, "OffHeapExecutionMemory", timeseries_db_df, "4. Peak ")  # off heap is disable by default in Spark
    ax.set_title("4. Local memory usage:")
    timeseries_db_df["process_memory_usage"].plot(ax=ax)  # We should put only 1 ylim, the last one
    plot_max_value(ax, timeseries_db_df["process_memory_usage"], convert_size)
    timeseries_db_df["user_memory_usage"].plot(ax=ax, ylim=(0, None))
    plot_max_value(ax, timeseries_db_df["user_memory_usage"], convert_size)
    ax_convert_size(ax)
    ax.legend()
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[3:5, 0])
    prepare_axis(ax)
    mmm_plot(ax, "totalOnHeapStorageMemory", timeseries_db_df, "5. ")
    # mmm_plot(ax, 'usedOnHeapStorageMemory', timeseries_db_df, '5. ')  # 0
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[3:5, 1])
    prepare_axis(ax)
    ax.set_title("6. Driver memory percentage:")
    timeseries_db_df["memoryUsedPct_driver"].plot(label="driver", ax=ax)
    plot_max_value(ax, timeseries_db_df["memoryUsedPct_driver"])
    # mmm_plot(ax, 'totalOffHeapStorageMemory', timeseries_db_df, '6. ')
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[5:7, 0])
    prepare_axis(ax)
    mmm_plot(ax, "ProcessTreePythonRSSMemory", timeseries_db_df, "7. Peak ")
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[5:7, 1])
    prepare_axis(ax)
    mmm_plot(ax, "ProcessTreePythonVMemory", timeseries_db_df, "8. Peak ")
    ax.set_xticklabels([])

    ax = fig.add_subplot(gs[7:9, 0])
    prepare_axis(ax)
    mmm_plot(ax, "JVMHeapMemory", timeseries_db_df, "9. Peak ")

    ax = fig.add_subplot(gs[7:9, 1])
    prepare_axis(ax)
    mmm_plot(ax, "JVMOffHeapMemory", timeseries_db_df, "10. Peak ")

    return fig


def plot_notebook(application: sparkmon.Application) -> None:
    """Plot for notebook."""
    display.clear_output(True)
    application.plot()
    plt.show()

    # To avoid memory leak
    plt.clf()
    plt.close()



# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""List of default callbacks."""
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt

from sparkmon import Application
from sparkmon.mlflow_utils import log_file


def plot_to_image(application: Application, path: str = "sparkmon.png") -> None:
    """Plot and save to image.

    Not compatible with live_plot_notebook().
    """
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    # To avoid the following error: "RuntimeError: main thread is not in main loop"
    backend = "agg"
    if matplotlib.get_backend() != backend:
        print("set backend")
        plt.switch_backend(backend)

    application.plot()
    plt.savefig(path)

    # To avoid memory leak
    plt.clf()
    plt.close()


def plot_to_mlflow(application: Application, path: str = "sparkmon/plot.png") -> None:
    """Log image to mlflow.

    Not compatible with live_plot_notebook().
    """
    # To avoid the following error: "RuntimeError: main thread is not in main loop"
    backend = "agg"
    if matplotlib.get_backend() != backend:
        print("set backend")
        plt.switch_backend(backend)

    application.plot()
    with log_file(path) as fp:
        plt.savefig(fp.name)

    # To avoid memory leak
    plt.clf()
    plt.close()


def log_timeseries_db_to_mlflow(application: Application, path: str = "sparkmon/timeseries.csv") -> None:
    """Log timeseries_db to mlflow."""
    timeseries_db_df = application.get_timeseries_db_df()
    with log_file(path) as fp:
        timeseries_db_df.to_csv(fp, index=False)


def log_tasks_to_mlflow(application: Application, path: str = "sparkmon/tasks.csv") -> None:
    """Log tasks to mlflow."""
    with log_file(path) as fp:
        application.get_tasks_df().to_csv(fp, index=False)


def log_stages_to_mlflow(application: Application, path: str = "sparkmon/stages.csv") -> None:
    """Log tasks to mlflow."""
    with log_file(path) as fp:
        application.stages_df.to_csv(fp, index=False)


def log_to_mlflow(application: Application) -> None:
    """Log to mlflow."""
    plot_to_mlflow(application)
    log_timeseries_db_to_mlflow(application)
    log_tasks_to_mlflow(application)
    log_stages_to_mlflow(application)



# Copyright (c) 2021 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Spark communication interface with its API, and managing historical API calls."""
import http
import warnings
from datetime import datetime
from typing import Any
from typing import Dict

import pandas as pd
import psutil
import requests
import urllib3
import urlpath
from pyspark.sql import SparkSession

import sparkmon
from sparkmon.utils import flatten_dict
from sparkmon.utils import get_memory_process
from sparkmon.utils import get_memory_user

API_APPLICATIONS_LINK = "api/v1/applications"
WEB_URL = "http://localhost:4040"


class Application:
    """This class is an helper to query Spark API and save historical."""

    def __init__(self, application_id: str, web_url: str = WEB_URL, debug: bool = False) -> None:
        """An application is define by the Spark UI link and an application id.

        :param application_id: Spark applicationId
        :param web_url: Spark REST API server
        :param debug: debug mode for development purposes, it store the row data
        """
        self.web_url = web_url
        self.application_id = application_id
        self.debug = debug

        self.executors_db: Dict[Any, Any] = {}  # The key is timestamp
        self.timeseries_db: Dict[Any, Any] = {}  # The key is timestamp
        self.stages_df: pd.DataFrame = pd.DataFrame()
        self.tasks_db: Dict[str, Dict[Any, Any]] = {}  # The key is stageId.attemptId

    def get_executors_info(self) -> pd.DataFrame:
        """Retrieve executors info."""
        executors_df = pd.read_json(
            urlpath.URL(
                self.web_url,
                API_APPLICATIONS_LINK,
                self.application_id,
                "executors",
            )
        )
        return executors_df

    def log_executors_info(self) -> None:
        """Add a new executor info in the dict database."""
        executors_df = self.get_executors_info()

        now = pd.to_datetime(datetime.now())
        if self.debug:
            self.executors_db[now] = executors_df  # Storing full row data

        self.timeseries_db[now] = self.parse_executors(executors_df)
        # Remark: local_memory_pct is a percentage, doesn't really work if you are running an kube
        self.timeseries_db[now]["local_memory_pct"] = psutil.virtual_memory()[2]  # Local machine memory usage
        self.timeseries_db[now]["process_memory_usage"] = get_memory_process()  # Local machine memory usage
        self.timeseries_db[now]["user_memory_usage"] = get_memory_user()  # Local machine memory usage

    def parse_db(self) -> None:
        """Re-parse the full executors_db, usefull if you change the parsing function, for development."""
        if not self.debug:
            print("sparkmon: Warning, parse_db should be used with debug=True")
        for t, executors_df in self.executors_db.items():
            self.timeseries_db[t].update(self.parse_executors(executors_df))

    def plot(self) -> None:
        """Plotting."""
        sparkmon.plot_timeseries(self.get_timeseries_db_df(), title=self.application_id)

    @staticmethod
    def parse_executors(executors_df: pd.DataFrame) -> Dict[Any, Any]:
        """Convert an executors DataFrame to a dictionnary.

        More spefically we aggregate metrix from all executors into one.
        """
        if len(executors_df) == 0:
            return {}

        memoryMetrics = executors_df["memoryMetrics"].dropna()
        memoryMetrics_df = pd.DataFrame.from_records(memoryMetrics, index=memoryMetrics.index)
        executors_df = executors_df.join(memoryMetrics_df)

        if "peakMemoryMetrics" in executors_df.columns:
            peakMemoryMetrics = executors_df["peakMemoryMetrics"].dropna()
            peakMemoryMetrics_df = pd.DataFrame.from_records(peakMemoryMetrics, index=peakMemoryMetrics.index)
            executors_df = executors_df.join(peakMemoryMetrics_df)

        executors_df["usedOnHeapStorageMemoryPct"] = (
            executors_df["usedOnHeapStorageMemory"] / executors_df["totalOnHeapStorageMemory"] * 100
        )
        executors_df["usedOffHeapStorageMemoryPct"] = (
            executors_df["usedOffHeapStorageMemory"] / executors_df["totalOffHeapStorageMemory"] * 100
        )
        executors_df["memoryUsedPct"] = executors_df["memoryUsed"] / executors_df["maxMemory"] * 100

        # Columns to aggregate
        cols = [
            "memoryUsedPct",
            "usedOnHeapStorageMemoryPct",
            "usedOffHeapStorageMemoryPct",
            "totalOnHeapStorageMemory",
            "totalOffHeapStorageMemory",
            "ProcessTreePythonVMemory",
            "ProcessTreePythonRSSMemory",
            "JVMHeapMemory",
            "JVMOffHeapMemory",
            "OffHeapExecutionMemory",
            "OnHeapExecutionMemory",
        ]

        # Filter columns that are not on the DataFrame
        cols = [col for col in cols if col in executors_df.columns]

        # Obtain aggregated values
        # Let's filter: "RuntimeWarning: Mean of empty slice"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            res = executors_df[cols].agg(["max", "mean", "median", "min"])

        # Convert to dict
        d: Dict[str, Any] = {
            f"{key}_{agg}": value for key, values in res.to_dict().items() for agg, value in values.items()
        }

        d["numActive"] = len(executors_df.query("isActive"))
        d["memoryUsed_sum"] = executors_df["memoryUsed"].sum()
        d["maxMemory_sum"] = executors_df["maxMemory"].sum()
        d["memoryUsed_sum_pct"] = executors_df["memoryUsed"].sum() / executors_df["maxMemory"].sum() * 100
        d["memoryUsedPct_driver"] = executors_df.iloc[0]["memoryUsedPct"]

        return d

    def log_stages(self) -> None:
        """Retrieve stages."""
        url = urlpath.URL(self.web_url, API_APPLICATIONS_LINK, self.application_id, "stages")
        self.stages_df = pd.read_json(url)

    def log_tasks(self) -> None:
        """Retrieve tasks."""
        for _, row in self.stages_df.iterrows():
            stage_uid = f"{row['stageId']}.{row['attemptId']}"

            # Initialize
            if self.tasks_db.get(stage_uid) is None:
                self.tasks_db[stage_uid] = {"tasks": None, "stage_last_status": None}

            # Don't query again what is done
            if self.tasks_db[stage_uid]["stage_last_status"] in ["COMPLETE", "SKIPPED", "FAILED"]:
                continue

            url = urlpath.URL(
                self.web_url,
                API_APPLICATIONS_LINK,
                self.application_id,
                "stages",
                str(row["stageId"]),
                str(row["attemptId"]),
            )

            stage_detail_r = requests.get(url)
            stage_detail = stage_detail_r.json()

            tasks = stage_detail["tasks"]

            # Flatten the dictionnary
            tasks = [flatten_dict(tasks[k]) for k in tasks.keys()]

            self.tasks_db[stage_uid]["tasks"] = tasks
            self.tasks_db[stage_uid]["stage_last_status"] = row["status"]

    def get_tasks_df(self) -> pd.DataFrame:
        """Return all tasks info into a DataFrame."""
        tasks_list = []
        # Iter stages
        for k in self.tasks_db.keys():
            tasks = self.tasks_db[k]["tasks"]
            # Iter tasks
            for t in tasks:
                t["stage_uid"] = k
                tasks_list.append(t)
        tasks_df = pd.DataFrame(tasks_list)
        return tasks_df

    def get_timeseries_db_df(self) -> pd.DataFrame:
        """Return timeseries_db info into a DataFrame."""
        timeseries_db_df = pd.DataFrame(self.timeseries_db).T
        return timeseries_db_df

    def log_all(self) -> None:
        """Updating all information."""
        try:
            self.log_executors_info()
            self.log_stages()
            self.log_tasks()
        except requests.exceptions.ConnectionError as e:
            print(f"sparkmon: Warning ConnectionError while updating all information: {e}")
        except http.client.RemoteDisconnected as e:
            print(f"sparkmon: Warning RemoteDisconnected while updating all information: {e}")
        except urllib3.exceptions.ProtocolError as e:
            print(f"sparkmon: Warning ProtocolError while updating all information: {e}")


def get_application_ids(web_url: str = WEB_URL) -> pd.DataFrame:
    """Retrieve available application id."""
    applications_df = pd.read_json(urlpath.URL(web_url, API_APPLICATIONS_LINK))
    return applications_df


def create_application_from_link(index: int = 0, web_url: str = WEB_URL) -> Application:
    """Create an Application.

    :param index: Application index in the application list
    :param web_url: Spark REST API server
    """
    applications_df = get_application_ids(web_url)
    application_id = applications_df["id"].iloc[index]

    application = Application(application_id, web_url)

    return application


def create_application_from_spark(spark: SparkSession) -> Application:
    """Create an Application from Spark Session."""
    web_url = spark.sparkContext.uiWebUrl
    application_id = spark.sparkContext.applicationId

    application = Application(application_id, web_url)

    return application
